import unittest2
import sys
import gzip


import chemfp
from chemfp.commandline import simsearch
from cStringIO import StringIO

SOFTWARE = ("chemfp/" + chemfp.__version__).encode("ascii")

import support

class SimsearchRunner(support.Runner):
    def verify_result(self, result):
        assert result[0] == b"#Simsearch/1", result[0]
class CountRunner(support.Runner):
    def verify_result(self, result):
        assert result[0] == b"#Count/1", result[0]

runner = SimsearchRunner(simsearch.main)
run = runner.run
run_split = runner.run_split
run_exit = runner.run_exit

count_runner = CountRunner(simsearch.main)
count_run_split = count_runner.run_split
count_run_exit = count_runner.run_exit

SIMPLE_FPS = support.fullpath("simple.fps")
SIMPLE_FPS_GZ = support.fullpath("simple.fps.gz")

DRUGS_FLUSH = support.fullpath("drugs.flush")

def run_split_stdin(input, cmdline, expect_length=None, source=SIMPLE_FPS):
    old_stdin = sys.stdin
    sys.stdin = StringIO(input)
    try:
        return run_split(cmdline, expect_length, source)
    finally:
        sys.stdin = old_stdin

def gzip_compress(s):
    f = StringIO()
    g = gzip.GzipFile(fileobj=f, mode="w")
    g.write(s)
    g.close()
    return f.getvalue()


# The values I get using gmpy are:
#    [(1.0, 'deadbeef'),
#     (0.95999999999999996, 'Deaf Beef'),
#     (0.83999999999999997, 'DEADdead'),
#     (0.23999999999999999, 'several'),
#     (0.041666666666666664, 'bit1'),
#     (0.040000000000000001, 'two_bits'),
#     (0.0, 'zeros')]

class TestOptions(unittest2.TestCase):
    def test_default(self):
        header, lines = run_split("--hex-query deadbeef -t 0.1", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=all threshold=0.1"})
        self.assertEquals(len(lines), 1, lines)
        fields = lines[0].split(b"\t")
        self.assertEquals(fields[:2], [b"4", b"Query1"])
        hits = list(zip(fields[2::2], fields[3::2]))
        hits.sort()
        self.assertEquals(hits, [(b"DEADdead", b"0.840"), (b"Deaf Beef", b"0.960"),
                                 (b"deadbeef", b"1.000"), (b'several', b'0.240')])

    def test_k_3(self):
        header, lines = run_split("--hex-query deadbeef -k 3 --threshold 0.8", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.8"})
        self.assertEquals(lines,
                          [b"3\tQuery1\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840"])

    def test_k_2(self):
        header, lines = run_split("--hex-query deadbeef -k 2 --threshold 0.9", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=2 threshold=0.9"})
        self.assertEquals(lines,
                          [b"2\tQuery1\tdeadbeef\t1.000\tDeaf Beef\t0.960"])

    def test_k_1(self):
        header, lines = run_split("--hex-query deadbeef -k 1 -t 0.0", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=1 threshold=0.0"})
        self.assertEquals(lines,
                          [b"1\tQuery1\tdeadbeef\t1.000"])

    def test_knearest_1(self):
        header, lines = run_split("--hex-query deadbeef --k-nearest 1", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=1 threshold=0.0"})
        self.assertEquals(lines,
                          [b"1\tQuery1\tdeadbeef\t1.000"])

    def test_k_0(self):
        header, lines = run_split("--hex-query deadbeef -k 0 -t 0.0", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEqual(header,
                         {b"#num_bits": b"32",
                          b"#software": SOFTWARE,
                          b"#type": b"Tanimoto k=0 threshold=0.0"})
        self.assertEqual(lines,
                         [b"0\tQuery1"])

    def test_k_10(self):
        # Asked for 10 but only 7 are available
        header, lines = run_split("--hex-query deadbeef -k 10 --threshold 0.0", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=10 threshold=0.0"})
        self.assertEquals(lines,
                          [b"7\tQuery1\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840\t"
                           b"several\t0.240\tbit1\t0.042\ttwo_bits\t0.040\tzeros\t0.000"])

    def test_knearest_all(self):
        header, lines = run_split("--hex-query deadbeef --k-nearest all", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=all threshold=0.0"})
        self.assertEquals(lines,
                          [b'7\tQuery1\tzeros\t0.000\tbit1\t0.042\ttwo_bits\t0.040\tseveral\t0.240\tdeadbeef\t1.000\tDEADdead\t0.840\tDeaf Beef\t0.960'])

    def test_threshold(self):
        header, lines = run_split("--hex-query deadbeef --threshold 0.9", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=all threshold=0.9"})
        self.assertEquals(lines,
                          [b"2\tQuery1\tdeadbeef\t1.000\tDeaf Beef\t0.960"])

    def test_threshold_and_k(self):
        header, lines = run_split("--hex-query deadbeef -t 0.9 -k 1", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=1 threshold=0.9"})
        self.assertEquals(lines,
                          [b"1\tQuery1\tdeadbeef\t1.000"])

    def test_threshold_and_k_all(self):
        header, lines = run_split("--hex-query deadbeef --threshold 0.9 --k-nearest all", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=all threshold=0.9"})
        self.assertEquals(lines,
                          [b"2\tQuery1\tdeadbeef\t1.000\tDeaf Beef\t0.960"])

    
    def test_stdin(self):
        header, lines = run_split_stdin(b"deadbeef\tspam\n", "", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.7"})
        self.assertEquals(lines,
                          [b"3\tspam\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840"])

    def test_stdin2(self):
        header, lines = run_split_stdin(b"deadbeef\tspam\nDEADBEEF\teggs\n",
                                        "-k 3 --threshold 0.6", 2, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.6"})
        self.assertEquals(lines,
                          [b"3\tspam\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840",
                           b"3\teggs\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840"])

    def test_stdin3(self):
        header, lines = run_split_stdin(b"deadbeef\tspam\n87654321\tcountdown\n",
                                        "-k 3 --threshold 0.9", 2, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.9"})
        self.assertEquals(lines,
                          [b"2\tspam\tdeadbeef\t1.000\tDeaf Beef\t0.960",
                           b"0\tcountdown"])
    def test_in_format_deprecated(self):
        s = gzip_compress(b"deadbeef\tspam\n")
        # You should use use "--query-format" instead of "--in".
        header, lines = run_split_stdin(s, "--in fps.gz", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.7"})
        self.assertEquals(lines,
                          [b"3\tspam\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840"])

    def test_query_format_stdin(self):
        s = gzip_compress(b"deadbeef\tspam\n")
        header, lines = run_split_stdin(s, "--query-format fps.gz", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.7"})
        self.assertEquals(lines,
                          [b"3\tspam\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840"])

    def test_query_format_fps_gzip_but_actually_fps(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS, "--query-format", "fps.gz"], SIMPLE_FPS)
        self.assertIn("Not a gzipped file", errmsg)

    def test_query_format_fps_but_actually_fps_gz(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS_GZ, "--query-format", "fps"], SIMPLE_FPS)
        self.assertIn("Line must end with a newline character", errmsg)
        self.assertIn("line 1", errmsg)

    def test_query_format_unknown(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS, "--query-format", "f2s"], SIMPLE_FPS)
        self.assertIn("Unsupported fingerprint format 'f2s'", errmsg)

    def test_query_format_compression_unknown(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS, "--query-format", "fps.33"], SIMPLE_FPS)
        self.assertIn("chemfp does not support compression type '33'", errmsg)

    def test_target_format_fps(self):
        header, lines = run_split_stdin(b"deadbeef\tspam\n", "--target-format fps", 1, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.7"})
        self.assertEquals(lines,
                          [b"3\tspam\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840"])

    def test_target_format_fps_gz(self):
        header, lines = run_split_stdin(b"deadbeef\tspam\n", "--target-format fps.gz", 1, SIMPLE_FPS_GZ)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.7"})
        self.assertEquals(lines,
                          [b"3\tspam\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840"])

    def test_target_format_fps_but_actually_fps_gz(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS, "--target-format", "fps"], SIMPLE_FPS_GZ)
        self.assertIn("Cannot parse targets file: Line must end with a newline character", errmsg)
        self.assertIn(u"line 1\n", errmsg)

    def test_target_format_flush(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS], DRUGS_FLUSH)
        self.assertEqual(errmsg,
                 "Simsearch cannot use flush files as input. Use chemfp_converters to convert it to FPS or FPB format.\n")
        
    def test_target_format_unknown(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS, "--target-format", "fp3"], SIMPLE_FPS_GZ)
        self.assertIn("Unsupported fingerprint format 'fp3'", errmsg)

    def test_target_format_compression_unknown(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS, "--target-format", "fps.33"], SIMPLE_FPS_GZ)
        self.assertIn("Cannot open targets file: chemfp does not support compression type '33'\n", errmsg)
        
def normalize_line(line):
    fields = line.split(b"\t")
    return int(fields[0]), fields[1], set((fields[i], fields[i+1]) for i in range(2, len(fields), 2))
def normalize_lines(lines):
    return [normalize_line(line) for line in lines]
        
class _AgainstSelf:
    def test_with_threshold(self):
        header, lines = run_split(
            ["--queries", SIMPLE_FPS, "--threshold", "0.8"] + self.extra_arg,
            7, SIMPLE_FPS)
        expected_lines = [b"0\tzeros",
                          b"1\tbit1\tbit1\t1.000",
                          b"1\ttwo_bits\ttwo_bits\t1.000",
                          b"1\tseveral\tseveral\t1.000",
                          b"3\tdeadbeef\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840",
                          b"3\tDEADdead\tDEADdead\t1.000\tdeadbeef\t0.840\tDeaf Beef\t0.808",
                          b"3\tDeaf Beef\tDeaf Beef\t1.000\tdeadbeef\t0.960\tDEADdead\t0.808"]
        # the output order is arbitary
        self.assertEquals(normalize_lines(lines), normalize_lines(expected_lines))

    def test_with_k_and_threshold(self):
        header, lines = run_split(
            ["--queries", SIMPLE_FPS, "-k", "3", "--threshold", "0.8"] + self.extra_arg,
            7, SIMPLE_FPS)
        # don't need to normalize; the output must be ordered from highest to lowest.
        self.assertEquals(lines,
                          [b"0\tzeros",
                           b"1\tbit1\tbit1\t1.000",
                           b"1\ttwo_bits\ttwo_bits\t1.000",
                           b"1\tseveral\tseveral\t1.000",
                           b"3\tdeadbeef\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840",
                           b"3\tDEADdead\tDEADdead\t1.000\tdeadbeef\t0.840\tDeaf Beef\t0.808",
                           b"3\tDeaf Beef\tDeaf Beef\t1.000\tdeadbeef\t0.960\tDEADdead\t0.808"])
        
    def test_with_count_and_threshold(self):
        header, lines = count_run_split(
            ["--queries", SIMPLE_FPS, "--threshold", "0.8", "--count"] + self.extra_arg,
            7, SIMPLE_FPS)
        self.assertEquals(lines,
                          [b"0\tzeros",
                           b"1\tbit1",
                           b"1\ttwo_bits",
                           b"1\tseveral",
                           b"3\tdeadbeef",
                           b"3\tDEADdead",
                           b"3\tDeaf Beef"])

    def test_with_threshold_0(self):
        header, lines = run_split(
            ["--queries", SIMPLE_FPS, "-k", "3", "--threshold", "0.0"] + self.extra_arg,
            7, SIMPLE_FPS)
        # The order is implementation dependent. Normalize before testing.
        if lines[0] == b"3\tzeros\tbit1\t0.000\ttwo_bits\t0.000\tzeros\t0.000":
            lines[0] = b"3\tzeros\tzeros\t0.000\tbit1\t0.000\ttwo_bits\t0.000"
        self.assertEquals(lines,
                          [b"3\tzeros\tzeros\t0.000\tbit1\t0.000\ttwo_bits\t0.000",
                           b"3\tbit1\tbit1\t1.000\ttwo_bits\t0.500\tseveral\t0.143",
                           b"3\ttwo_bits\ttwo_bits\t1.000\tbit1\t0.500\tseveral\t0.286",
                           b"3\tseveral\tseveral\t1.000\ttwo_bits\t0.286\tDEADdead\t0.261",
                           b"3\tdeadbeef\tdeadbeef\t1.000\tDeaf Beef\t0.960\tDEADdead\t0.840",
                           b"3\tDEADdead\tDEADdead\t1.000\tdeadbeef\t0.840\tDeaf Beef\t0.808",
                           b"3\tDeaf Beef\tDeaf Beef\t1.000\tdeadbeef\t0.960\tDEADdead\t0.808"])

    def test_with_count_and_threshold_0(self):
        header, lines = count_run_split(
            ["--queries", SIMPLE_FPS, "--threshold", "0.0", "--count"] + self.extra_arg,
            7, SIMPLE_FPS)
        self.assertEquals(lines,
                          [b"7\tzeros",
                           b"7\tbit1",
                           b"7\ttwo_bits",
                           b"7\tseveral",
                           b"7\tdeadbeef",
                           b"7\tDEADdead",
                           b"7\tDeaf Beef"])

class TestAgainstSelf(unittest2.TestCase, _AgainstSelf):
    extra_arg = []

class TestAgainstSelfInMemory(unittest2.TestCase, _AgainstSelf):
    extra_arg = ["--memory"]

class TestAgainstSelfFileScan(unittest2.TestCase, _AgainstSelf):
    extra_arg = ["--scan"]


class TestNxN(unittest2.TestCase):
    def test_default(self):
        header, lines = run_split("--NxN", 7, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.7 NxN=full"})
        self.assertEquals(len(lines), 7, lines)
        self.assertEquals(lines, [b'0\tzeros',
                                  b'0\tbit1',
                                  b'0\ttwo_bits',
                                  b'0\tseveral',
                                  b'2\tdeadbeef\tDeaf Beef\t0.960\tDEADdead\t0.840',
                                  b'2\tDEADdead\tdeadbeef\t0.840\tDeaf Beef\t0.808',
                                  b'2\tDeaf Beef\tdeadbeef\t0.960\tDEADdead\t0.808'])
    def test_specify_default_values(self):
        header, lines = run_split("--NxN -k 3 --threshold 0.7", 7, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=3 threshold=0.7 NxN=full"})
        self.assertEquals(len(lines), 7, lines)
        self.assertEquals(lines, [b'0\tzeros',
                                  b'0\tbit1',
                                  b'0\ttwo_bits',
                                  b'0\tseveral',
                                  b'2\tdeadbeef\tDeaf Beef\t0.960\tDEADdead\t0.840',
                                  b'2\tDEADdead\tdeadbeef\t0.840\tDeaf Beef\t0.808',
                                  b'2\tDeaf Beef\tdeadbeef\t0.960\tDEADdead\t0.808'])

    def test_k_2(self):
        # This sets the theshold to 0.0
        header, lines = run_split("--NxN -k 2 ", 7, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=2 threshold=0.0 NxN=full"})
        self.assertEquals(len(lines), 7, lines)
        self.assertEquals(lines, [b'2\tzeros\tbit1\t0.000\ttwo_bits\t0.000',
                                  b'2\tbit1\ttwo_bits\t0.500\tseveral\t0.143',
                                  b'2\ttwo_bits\tbit1\t0.500\tseveral\t0.286',
                                  b'2\tseveral\ttwo_bits\t0.286\tDEADdead\t0.261',
                                  b'2\tdeadbeef\tDeaf Beef\t0.960\tDEADdead\t0.840',
                                  b'2\tDEADdead\tdeadbeef\t0.840\tDeaf Beef\t0.808',
                                  b'2\tDeaf Beef\tdeadbeef\t0.960\tDEADdead\t0.808'])

    def test_threshold(self):
        header, lines = run_split("--NxN --threshold 0.5 ", 7, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=all threshold=0.5 NxN=full"})
        self.assertEquals(len(lines), 7, lines)
        self.assertEquals(lines, [b'0\tzeros',
                                  b'1\tbit1\ttwo_bits\t0.500',
                                  b'1\ttwo_bits\tbit1\t0.500',
                                  b'0\tseveral',
                                  # The order here is implementation dependent...
                                  b'2\tdeadbeef\tDeaf Beef\t0.960\tDEADdead\t0.840',
                                  b'2\tDEADdead\tdeadbeef\t0.840\tDeaf Beef\t0.808',
                                  #'2\tDeaf Beef\tdeadbeef\t0.960\tDEADdead\t0.808',
                                  b'2\tDeaf Beef\tDEADdead\t0.808\tdeadbeef\t0.960',
            ])

    def test_count_with_threshold(self):
        header, lines = count_run_split("--NxN --count --threshold 0.5 ", 7, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Count threshold=0.5 NxN=full"})
        self.assertEquals(len(lines), 7, lines)
        self.assertEquals(lines, [b'0\tzeros',
                                  b'1\tbit1',
                                  b'1\ttwo_bits',
                                  b'0\tseveral',
                                  b'2\tdeadbeef',
                                  b'2\tDEADdead',
                                  b'2\tDeaf Beef',
            ])

    def test_count_with_default_threshold(self):
        header, lines = count_run_split("--NxN --count", 7, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Count threshold=0.7 NxN=full"})
        self.assertEquals(len(lines), 7, lines)
        self.assertEquals(lines, [b'0\tzeros',
                                  b'0\tbit1',
                                  b'0\ttwo_bits',
                                  b'0\tseveral',
                                  b'2\tdeadbeef',
                                  b'2\tDEADdead',
                                  b'2\tDeaf Beef',
            ])


        
    def test_threshold_with_low_batch_size(self):
        header, lines = run_split("--NxN --threshold 0.5 --batch-size 1", 7, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                          b"#software": SOFTWARE,
                          b"#type": b"Tanimoto k=all threshold=0.5 NxN=full"})
        self.assertEquals(len(lines), 7, lines)
        self.assertEquals(lines, [b'0\tzeros',
                                  b'1\tbit1\ttwo_bits\t0.500',
                                  b'1\ttwo_bits\tbit1\t0.500',
                                  b'0\tseveral',
                                  # The order here is implementation dependent...
                                  b'2\tdeadbeef\tDeaf Beef\t0.960\tDEADdead\t0.840',
                                  b'2\tDEADdead\tdeadbeef\t0.840\tDeaf Beef\t0.808',
                                  #'2\tDeaf Beef\tdeadbeef\t0.960\tDEADdead\t0.808',
                                  b'2\tDeaf Beef\tDEADdead\t0.808\tdeadbeef\t0.960',
            ])
    def test_knearest_with_low_batch_size(self):
        # This is the same as test_k_2 but with a batch-size of 1.
        # This tests a bug where I wasn't incrementing the offset
        # to the start of each batch location in the results.
        header, lines = run_split("--NxN -k 2 --batch-size 1", 7, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Tanimoto k=2 threshold=0.0 NxN=full"})
        self.assertEquals(len(lines), 7, lines)
        self.assertEquals(lines, [b'2\tzeros\tbit1\t0.000\ttwo_bits\t0.000',
                                  b'2\tbit1\ttwo_bits\t0.500\tseveral\t0.143',
                                  b'2\ttwo_bits\tbit1\t0.500\tseveral\t0.286',
                                  b'2\tseveral\ttwo_bits\t0.286\tDEADdead\t0.261',
                                  b'2\tdeadbeef\tDeaf Beef\t0.960\tDEADdead\t0.840',
                                  b'2\tDEADdead\tdeadbeef\t0.840\tDeaf Beef\t0.808',
                                  b'2\tDeaf Beef\tdeadbeef\t0.960\tDEADdead\t0.808'])
        
    def test_count_with_low_batch_size(self):
        header, lines = count_run_split("--NxN --count --batch-size 1", 7, SIMPLE_FPS)
        self.assertIn(b"simple.fps", header.pop(b"#targets"))
        self.assertEquals(header,
                          {b"#num_bits": b"32",
                           b"#software": SOFTWARE,
                           b"#type": b"Count threshold=0.7 NxN=full"})
        self.assertEquals(len(lines), 7, lines)
        self.assertEquals(lines, [b'0\tzeros',
                                  b'0\tbit1',
                                  b'0\ttwo_bits',
                                  b'0\tseveral',
                                  b'2\tdeadbeef',
                                  b'2\tDEADdead',
                                  b'2\tDeaf Beef',
            ])

        
        
        
class TestCompatibility(unittest2.TestCase):
    def test_incompatible_fingerprint(self):
        errmsg = run_exit(["--hex-query", "dead"], SIMPLE_FPS)
        self.assertIn("error: query fingerprint contains 2 bytes but", errmsg)
        self.assertIn("simple.fps", errmsg)
        self.assertIn("has 4 byte fingerprints", errmsg)

    def test_targets_is_not_an_fps_file(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS])
        self.assertIn("Cannot open targets file:", errmsg)
        self.assertIn("Unable to determine fingerprint format type from", errmsg)
        self.assertIn("pubchem.sdf'\n", errmsg)

    def test_targets_does_not_exist(self):
        errmsg = run_exit(["--queries", SIMPLE_FPS], "/this/file/does_not_exist_t")
        self.assertIn("Cannot open targets file:", errmsg)
        self.assertIn("No such file or directory", errmsg) # Mac specific?
        self.assertIn("does_not_exist_t", errmsg)
        self.assertTrue(errmsg.endswith("'\n"))

    def test_queries_is_not_an_fps_file(self):
        errmsg = run_exit(["--queries", support.PUBCHEM_SDF], SIMPLE_FPS)
        self.assertIn("Cannot open queries file:", errmsg)
        self.assertIn("Unable to determine fingerprint format type from ", errmsg)
        self.assertIn("pubchem.sdf'\n", errmsg)

    def test_queries_does_not_exist(self):
        errmsg = run_exit(["--queries", "/this/file/does_not_exist_q"], SIMPLE_FPS)
        self.assertIn("Cannot open queries file:", errmsg)
        self.assertIn("No such file or directory", errmsg) # Mac specific?
        self.assertIn("does_not_exist_q'\n", errmsg)

class TestCommandlineErrors(unittest2.TestCase):
    def test_mix_count_and_knearest(self):
        errmsg = count_run_exit("--count --hex-query beefcafe --k-nearest 4", SIMPLE_FPS)
        self.assertIn("--count search does not support --k-nearest", errmsg)
        
    def test_negative_k(self):
        errmsg = run_exit("--hex-query beefcafe -k -1", SIMPLE_FPS)
        self.assertIn("--k-nearest must be non-negative or 'all'", errmsg)

    def test_negative_threshold(self):
        errmsg = run_exit("--hex-query beefcafe --threshold -0.1", SIMPLE_FPS)
        self.assertIn("--threshold must be between 0.0 and 1.0, inclusive", errmsg)
        errmsg = run_exit("--hex-query beefcafe --threshold -1.0", SIMPLE_FPS)
        self.assertIn("--threshold must be between 0.0 and 1.0, inclusive", errmsg)

    def test_too_large_threshold(self):
        errmsg = run_exit("--hex-query beefcafe --threshold 1.1", SIMPLE_FPS)
        self.assertIn("--threshold must be between 0.0 and 1.0, inclusive", errmsg)

    def test_non_positive_batch_size(self):
        errmsg = run_exit("--hex-query beefcafe --batch-size 0", SIMPLE_FPS)
        self.assertIn("--batch-size must be positive", errmsg)
        errmsg = run_exit("--hex-query beefcafe --batch-size -1", SIMPLE_FPS)
        self.assertIn("--batch-size must be positive", errmsg)

    def test_NxN_with_scan(self):
        errmsg = run_exit("--NxN --scan", SIMPLE_FPS)
        self.assertIn("Cannot specify --scan with an --NxN search", errmsg)

    def test_NxN_with_hex_query(self):
        errmsg = run_exit("--NxN --hex-query feedfeed", SIMPLE_FPS)
        self.assertIn("Cannot specify --hex-query with an --NxN search", errmsg)
        
    def test_NxN_with_queries(self):
        errmsg = run_exit("--NxN --queries ignored", SIMPLE_FPS)
        self.assertIn("Cannot specify --queries with an --NxN search", errmsg)

    def test_scan_with_memory(self):
        errmsg = run_exit("--scan --memory", SIMPLE_FPS)
        self.assertIn("Cannot specify both --scan and --memory", errmsg)
        
    def test_hex_query_with_queries(self):
        errmsg = run_exit("--hex-query faceb00c --queries not_important", SIMPLE_FPS)
        self.assertIn("Can only specify at most one of --query, --hex-query, or --queries", errmsg)

    def test_query_with_queries(self):
        errmsg = run_exit("--query c1ccccc1 --queries not_important", SIMPLE_FPS)
        self.assertIn("Can only specify at most one of --query, --hex-query, or --queries", errmsg)

    def test_hex_query_with_query(self):
        errmsg = run_exit("--hex-query faceb00c --query CO", SIMPLE_FPS)
        self.assertIn("Can only specify at most one of --query, --hex-query, or --queries", errmsg)

    def test_all_queries(self):
        errmsg = run_exit("--hex-query faceb00c --query CO --queries not_important", SIMPLE_FPS)
        self.assertIn("Can only specify at most one of --query, --hex-query, or --queries", errmsg)

    def test_hex_query_with_bad_character(self):
        errmsg = run_exit("--hex-query faceb00k", SIMPLE_FPS)
        self.assertIn("--hex-query is not a hex string: Non-hexadecimal digit found", errmsg)
        
    def test_hex_query_with_bad_length(self):
        errmsg = run_exit("--hex-query deadbeef2", SIMPLE_FPS)
        self.assertIn("--hex-query is not a hex string: Odd-length string", errmsg)

    def test_query_id_with_bad_character(self):
        for (bad_id, name) in (("A\tB", b"tab"), ("C\nD", b"newline"),
                               ("E\rF", b"control-return"), ("G\0H", b"NUL")):
            errmsg = run_exit(["--hex-query", "abcd1234", "--query-id", bad_id], SIMPLE_FPS)
            self.assertIn("--query-id must not contain the %s character" % (name.decode("ascii"),), errmsg)

class _TestFPSParseErrors(object):
    def _test_missing_newline(self, data):
        assert data[-1:] != "\n"
        filename = support.get_tmpfile(self, "missing_terminal_newline.fps")
        with open(filename, "w") as outfile:
            outfile.write(data)
        errmsg = self._run_exit(filename)
        self.assertNotIn("Traceback (most recent call last)", errmsg)
        self.assertIn("Line must end with a newline character", errmsg)
        return errmsg

    def test_missing_FPS1_newline(self):
        errmsg = self._test_missing_newline("#FPS1")
        self.assertIn("line 1\n", errmsg)
        
    def test_missing_num_bytes_newline(self):
        errmsg = self._test_missing_newline("#FPS1\n#num_bytes=2")
        self.assertIn("line 2\n", errmsg)
    
    def test_missing_first_fp_newline(self):
        errmsg = self._test_missing_newline("12ef\tspam")
        self.assertIn("line 1\n", errmsg)
        
    def test_missing_first_fp_newline_with_FPS(self):
        errmsg = self._test_missing_newline("#FPS1\n12ef\tspam")
        self.assertIn("line 2\n", errmsg)
    
    def test_missing_first_fp_newline_with_num_bytes(self):
        errmsg = self._test_missing_newline("#FPS1\n#num_bytes=2\n12ef\tspam")
        self.assertIn("line 3\n", errmsg)

class TestTargetParseErrorsTanimoto(unittest2.TestCase, _TestFPSParseErrors):
    def _run_exit(self, filename):
        return run_exit(["--hex", "12ef"], filename)

class TestQueryParseErrors(unittest2.TestCase, _TestFPSParseErrors):
    def _run_exit(self, query_filename):
        target_filename = support.get_tmpfile(self, "targets.fps")
        with open(target_filename, "w") as outfile:
            outfile.write("1234\ttarget\n")
        return run_exit(["--queries", query_filename], target_filename)
        

        
    ## def test_missing_terminal_newline_has_num_bits(self):
    ##     filename = support.get_tmpfile(self, "missing_terminal_newline2.fps")
    ##     with open(filename, "w") as outfile:
    ##         outfile.write("#num_bits=16\n12ef\ttest1\n34cd\ttest")
    ##     errmsg = run_exit(["--hex", "12cd"], filename)
    ##     print(errmsg)

    ## def test_blank_line(self):
    ##     pass
    ## def test_missing_fingerprint(self):
    ##     pass

    
class TestWithEmptyFile(unittest2.TestCase):
    def _make_empty_file(self):
        empty_file = support.get_tmpfile(self, "empty.fps")
        open(empty_file, "wb").close()
        return empty_file
        
    def test_count_tanimoto_hits(self):
        # empty queries
        empty = self._make_empty_file()
        header, lines = count_run_split(["--count", "--queries", empty], 0, SIMPLE_FPS)
        
        # empty targets
        header, lines = count_run_split(["--count", "--queries", SIMPLE_FPS], 7, empty)
        for line in lines:
            self.assertTrue(line.startswith("0\t"), line)
        
    def test_threshold_tanimoto_search(self):
        # empty queries
        empty = self._make_empty_file()
        header, lines = run_split(["--threshold", "0.0", "--queries", empty], 0, SIMPLE_FPS)
        
        # empty targets
        header, lines = run_split(["--threshold", "0.0", "--queries", SIMPLE_FPS], 7, empty)
        for line in lines:
            self.assertTrue(line.startswith("0\t"), line)
        
    def test_knearest_tanimoto_search(self):
        # empty queries
        empty = self._make_empty_file()
        header, lines = run_split(["--k", "1", "--queries", empty], 0, SIMPLE_FPS)
        
        # empty targets
        header, lines = run_split(["--k", "1", "--queries", SIMPLE_FPS], 7, empty)
        for line in lines:
            self.assertTrue(line.startswith("0\t"), line)
        
    def test_count_tanimoto_hits_symmetric(self):
        empty = self._make_empty_file()
        header, lines = count_run_split(["--count", "--NxN"], 0, empty)
        
    def test_threshold_tanimoto_search_symmetric(self):
        empty = self._make_empty_file()
        header, lines = run_split(["--threshold", "0.0", "--NxN"], 0, empty)
        
    def test_knearest_tanimoto_search_symmetric(self):
        empty = self._make_empty_file()
        header, lines = run_split(["--k", "1", "--NxN"], 0, empty)

    
if __name__ == "__main__":
    unittest2.main()



from __future__ import with_statement

import os
import tempfile
import shutil
from cStringIO import StringIO

import unittest2

import chemfp
from chemfp.bitops import hex_decode

import support

def _tmpdir(testcase):
    dirname = tempfile.mkdtemp()
    testcase.addCleanup(shutil.rmtree, dirname)
    return dirname

# These are the same for FPS and FPB output formats.
# (1.3 doesn't support the FPB format. These were
# backported from 3.0.)
class OutputMixin(object):
    test_filename = None
    def _get_writer(self, **kwargs):
        d = self.test_kwargs.copy()
        d.update(kwargs)
        self.dirname = dirname = _tmpdir(self)
        #self.dirname = dirname = "."
        self.filename = filename = os.path.join(dirname, self.test_filename)
        return chemfp.open_fingerprint_writer(filename, **d)

    def test_save_empty_file(self):
        writer = self._get_writer()
        self.assertFalse(writer.closed)
        writer.close()
        self.assertTrue(writer.closed)

        for id, fp in chemfp.open(self.filename):
            raise AssertionError("Should be empty")
        
    
    def test_save_basic(self):
        writer = self._get_writer()
        writer.write_fingerprint("ABC", b"\r\t\n\0")
        writer.close()

        reader = iter(chemfp.open(self.filename))
        id, fp = next(reader)
        self.assertEqual(id, "ABC")
        self.assertEqual(fp, b"\r\t\n\0")
        with self.assertRaisesRegexp(StopIteration, ""):
            next(reader)

    def test_write_iterator_without_metadata(self):
        fps = [("AB", b"1234")]
        
        x = self._get_writer()
        with x as writer:
            writer.write_fingerprints(fps)

        reader = chemfp.open(self.filename)
        self.assertEqual(reader.metadata.type, None)
        self.assertEqual(reader.metadata.num_bits, 32)
        self.assertEqual(reader.metadata.num_bytes, 4)
        data = list(reader)
        self.assertEqual(data, [("AB", b"1234")])
        
    def test_write_iterator_with_metadata(self):
        fps = chemfp.FingerprintIterator(chemfp.Metadata(type="Spam/1", num_bytes=4), [("AB", b"1234")])
        
        x = self._get_writer(metadata=fps.metadata)
        with x as writer:
            self.assertFalse(x.closed)
            writer.write_fingerprints(fps)
        self.assertTrue(x.closed)

        reader = chemfp.open(self.filename)
        self.assertEqual(reader.metadata.type, "Spam/1")
        self.assertEqual(reader.metadata.num_bits, 32)
        self.assertEqual(reader.metadata.num_bytes, 4)
        data = list(reader)
        self.assertEqual(data, [("AB", b"1234")])

    def test_write_with_unusual_bits(self):
        with self._get_writer(metadata=chemfp.Metadata(num_bits=63)) as writer:
            writer.write_fingerprint("first", b"        ")
            writer.write_fingerprint("second", b"ZXCVBNM<")

        reader = chemfp.open(self.filename)
        self.assertEqual(reader.metadata.num_bytes, 8)
        self.assertEqual(reader.metadata.num_bits, 63)
        values = list(reader)
        self.assertEqual(values, [("first", b"        "),
                                  ("second", b"ZXCVBNM<")])

    def test_force_format_fps(self):
        with self._get_writer(format="fps") as writer:
            writer.write_fingerprint("first", hex_decode("deadbeef"))
        text = open(self.filename).read()
        self.assertEqual(text, "#FPS1\ndeadbeef\tfirst\n")  # though capitalization shouldn't matter

    def test_force_format_fps_gz(self):
        with self._get_writer(format="fps.gz") as writer:
            writer.write_fingerprint("first", hex_decode("deadbeef"))
        import gzip
        text = gzip.open(self.filename).read()
        self.assertEqual(text, b"#FPS1\ndeadbeef\tfirst\n")  # though capitalization shouldn't matter


class TestFPSOutput(unittest2.TestCase, OutputMixin):
    test_filename = "scratch.fps"
    test_kwargs = {}

    def test_save_id_with_tab(self):
        # The binary writer doesn't worry about special characters
        fps = [("Hello!", b"FFFF"), ("A\tB", b"1234")]
        with self._get_writer() as writer:
            with self.assertRaisesRegexp(
                    ValueError,
                    r"Unable to write an identifier containing a tab: 'A\\tB'.*scratch.fps', line 2, record #2"):
                writer.write_fingerprints(fps)

    def test_save_id_with_newline(self):
        # The binary writer doesn't worry about special characters
        fps = [("Hello!", b"FFFF"), ("AB", b"1234"),  ("A\nB", b"1234")]
        with self._get_writer() as writer:
            with self.assertRaisesRegexp(
                    ValueError,
                    r"Unable to write an identifier containing a newline: 'A\\nB'.*scratch.fps', line 3, record #3"):
                writer.write_fingerprints(fps)

    def test_save_id_with_empty_id(self):
        # The binary writer doesn't worry about empty ids
        fps = [("Hello!", b"FFFF"), ("", b"1234")]
        with self._get_writer() as writer:
            with self.assertRaisesRegexp(
                    ValueError,
                    r"Unable to write a fingerprint with an empty identifier.*scratch.fps', line 2, record #2"):
                writer.write_fingerprints(fps)


class UserWriter(object):
    def __init__(self):
        self.value = b""
    def write(self, text):
        self.value += text
    
class TestUserDefinedWriter(unittest2.TestCase):
    def test_user_writer(self):
        f = UserWriter()
        writer = chemfp.open_fingerprint_writer(f)
        writer.write_fingerprint("first", b"AA")
        writer.close()
        self.assertEqual(f.value, b"#FPS1\n4141\tfirst\n")
        
    def test_user_writer_must_implement_write(self):
        class NotAWriter(object):
            pass
        f = NotAWriter()
        with self.assertRaisesRegexp(ValueError, "Unknown destination type.*NotAWriter"):
            chemfp.open_fingerprint_writer(f)


class TestWriterLocation(unittest2.TestCase):
    def test_write_empty(self):
        f = StringIO()
        with chemfp.open_fingerprint_writer(f) as writer:
            self.assertEqual(writer.location.lineno, 1)
        self.assertEqual(f.getvalue(), b"#FPS1\n")
            
    def test_write_metadata_bytes(self):
        f = StringIO()
        with chemfp.open_fingerprint_writer(f, metadata=chemfp.Metadata(num_bytes=8)) as writer:
            self.assertEqual(writer.location.lineno, 2)
        self.assertEqual(f.getvalue(), b"#FPS1\n#num_bits=64\n")
            
    def test_write_metadata_bits(self):
        f = StringIO()
        with chemfp.open_fingerprint_writer(f, metadata=chemfp.Metadata(num_bits=60)) as writer:
            self.assertEqual(writer.location.lineno, 2)
        self.assertEqual(f.getvalue(), b"#FPS1\n#num_bits=60\n")
            
    def test_write_complex_metadata(self):
        f = StringIO()
        metadata = chemfp.Metadata(
            type = "blah/1",
            num_bytes = 16,
            sources = [u"G\x00F6teborg", u"Trollh\00E4ttan", "Miami"],
            )
            
        with chemfp.open_fingerprint_writer(f, metadata=metadata) as writer:
            self.assertEqual(writer.location.lineno, 6)
        self.assertEqual(
            f.getvalue(),
            b"#FPS1\n"
            b"#num_bits=128\n"
            b"#type=blah/1\n"
            b"#source=G\x00F6teborg\n"
            b"#source=Trollh\x00E4ttan\n"
            b"#source=Miami\n"
            )

    def test_write_fingerprint(self):
        f = StringIO()
        with chemfp.open_fingerprint_writer(f, chemfp.Metadata(num_bytes=2)) as writer:
            self.assertEqual(writer.location.lineno, 2)
            self.assertEqual(writer.location.recno, 0)
            self.assertEqual(writer.location.output_recno, 0)

            writer.write_fingerprint("first", b"AB")
            self.assertEqual(writer.location.recno, 1)
            self.assertEqual(writer.location.output_recno, 1)

            try:
                writer.write_fingerprint("", b"AB")
            except chemfp.ParseError as err:
                self.assertEqual("Unable to write a fingerprint with an empty identifier", err.msg)
                self.assertIs(err.location, writer.location)
                self.assertEqual(writer.location.recno, 2)
                self.assertEqual(writer.location.output_recno, 1)
            else:
                raise AssertionError
            
            writer.write_fingerprint("second", b"BC")
            self.assertEqual(writer.location.recno, 3)
            self.assertEqual(writer.location.output_recno, 2)
            
            try:
                writer.write_fingerprint("q\tw", b"CD")
            except chemfp.ParseError as err:
                self.assertIn("Unable to write an identifier containing a tab", err.msg)
            else:
                raise AssertionError
                
            try:
                writer.write_fingerprint("q\nw", b"EF")
            except chemfp.ParseError as err:
                self.assertIn("Unable to write an identifier containing a newline", err.msg)
            else:
                raise AssertionError
            
            writer.write_fingerprint("third", b"GH")

            self.assertEqual(writer.location.recno, 6)
            self.assertEqual(writer.location.output_recno, 3)
            self.assertEqual(writer.location.lineno, 5)
        
        self.assertEqual(f.getvalue(),
                b"#FPS1\n"
                b"#num_bits=16\n"
                b"4142\tfirst\n"
                b"4243\tsecond\n"
                b"4748\tthird\n"
                )
        
    def test_write_fingerprint_ignore_errors(self):
        f = StringIO()
        with chemfp.open_fingerprint_writer(f, chemfp.Metadata(num_bytes=2), errors="ignore") as writer:
            self.assertEqual(writer.location.lineno, 2)
            self.assertEqual(writer.location.recno, 0)
            self.assertEqual(writer.location.output_recno, 0)

            writer.write_fingerprint("first", b"AB")
            self.assertEqual(writer.location.recno, 1)
            self.assertEqual(writer.location.output_recno, 1)

            writer.write_fingerprint("", b"AB")
            self.assertEqual(writer.location.recno, 2)
            self.assertEqual(writer.location.output_recno, 1)
            
            writer.write_fingerprint("second", b"BC")
            self.assertEqual(writer.location.recno, 3)
            self.assertEqual(writer.location.output_recno, 2)
            
            writer.write_fingerprint("q\tw", b"CD")
                
            writer.write_fingerprint("q\nw", b"EF")
            
            writer.write_fingerprint("third", b"GH")

            self.assertEqual(writer.location.recno, 6)
            self.assertEqual(writer.location.output_recno, 3)
            self.assertEqual(writer.location.lineno, 5)

        self.assertEqual(f.getvalue(),
                b"#FPS1\n"
                b"#num_bits=16\n"
                b"4142\tfirst\n"
                b"4243\tsecond\n"
                b"4748\tthird\n"
                )
        
    def test_write_many_fingerprints(self):
        f = StringIO()
        with chemfp.open_fingerprint_writer(f, chemfp.Metadata(num_bytes=2)) as writer:
            writer.write_fingerprints( [("first", b"AB"),
                                        ("second", b"BC")] )
            self.assertEqual(writer.location.lineno, 4)
            self.assertEqual(writer.location.recno, 2)
            self.assertEqual(writer.location.output_recno, 2)

            writer.write_fingerprints( [] )
            
            try:
                writer.write_fingerprints( [("third", b"CD"),
                                            ("q\tw", b"DE"),
                                            ("fourth", b"EF")] )
            except chemfp.ParseError as err:
                self.assertIn("Unable to write an identifier containing a tab", err.msg)
                self.assertIs(writer.location, err.location)
                self.assertEqual(err.location.lineno, 5)
                self.assertEqual(err.location.recno, 4)
                self.assertEqual(err.location.output_recno, 3)
            else:
                raise AssertionError(f.getvalue())

            try:
                writer.write_fingerprints( [("fifth", b"FG"),
                                            ("q\ne", b"GH"),
                                            ("sixth", b"HI")] )
            except chemfp.ParseError as err:
                self.assertIn("Unable to write an identifier containing a newline", err.msg)
                self.assertIs(writer.location, err.location)
                self.assertEqual(err.location.lineno, 6)
                self.assertEqual(err.location.recno, 6)
                self.assertEqual(err.location.output_recno, 4)
            else:
                raise AssertionError(f.getvalue())

            try:
                writer.write_fingerprints( [("seventh", b"IJ"),
                                            ("", b"GH"),
                                            ("eighth", b"JK")] )
            except chemfp.ParseError as err:
                self.assertIn("Unable to write a fingerprint with an empty identifier", err.msg)
                self.assertIs(writer.location, err.location)
                self.assertEqual(err.location.lineno, 7)
                self.assertEqual(err.location.recno, 8)
                self.assertEqual(err.location.output_recno, 5)
            else:
                raise AssertionError(f.getvalue())

        self.assertEqual(
            f.getvalue(),
            b"#FPS1\n"
            b"#num_bits=16\n"
            b"4142\tfirst\n"
            b"4243\tsecond\n"
            b"4344\tthird\n"
            b"4647\tfifth\n"
            b"494a\tseventh\n")

    def test_write_many_fingerprints_ignore(self):
        f = StringIO()
        with chemfp.open_fingerprint_writer(f, chemfp.Metadata(num_bytes=2), errors="ignore") as writer:
            writer.write_fingerprints( [("first", b"AB"),
                                        ("second", b"BC")] )
            self.assertEqual(writer.location.lineno, 4)
            self.assertEqual(writer.location.recno, 2)
            self.assertEqual(writer.location.output_recno, 2)

            writer.write_fingerprints( [] )
            
            writer.write_fingerprints( [("third", b"CD"),
                                        ("q\tw", b"DE"),
                                        ("fourth", b"EF")] )
            self.assertEqual(writer.location.lineno, 6)
            self.assertEqual(writer.location.recno, 5)
            self.assertEqual(writer.location.output_recno, 4)

            writer.write_fingerprints( [("fifth", b"FG"),
                                        ("q\ne", b"GH"),
                                        ("sixth", b"HI")] )
            self.assertEqual(writer.location.lineno, 8)
            self.assertEqual(writer.location.recno, 8)
            self.assertEqual(writer.location.output_recno, 6)

            writer.write_fingerprints( [("seventh", b"IJ"),
                                        ("", b"GH"),
                                        ("eighth", b"JK")] )
            self.assertEqual(writer.location.lineno, 10)
            self.assertEqual(writer.location.recno, 11)
            self.assertEqual(writer.location.output_recno, 8)

        self.assertEqual(
            f.getvalue(),
            b"#FPS1\n"
            b"#num_bits=16\n"
            b"4142\tfirst\n"
            b"4243\tsecond\n"
            b"4344\tthird\n"
            b"4546\tfourth\n"
            b"4647\tfifth\n"
            b"4849\tsixth\n"
            b"494a\tseventh\n"
            b"4a4b\teighth\n")

class TestErrorHandling(unittest2.TestCase):
    def test_unknown_format_name(self):
        with self.assertRaisesRegexp(ValueError, "Unable to determine fingerprint format type from 'blah.zztop"):
            chemfp.open_fingerprint_writer(support.get_tmpfile(self, "blah.zztop"))
            
    def test_unknown_format_name(self):
        with self.assertRaisesRegexp(ValueError, "Unsupported output fingerprint format 'zztop'"):
            chemfp.open_fingerprint_writer(support.get_tmpfile(self, "blah.fps"), format="zztop")
        
if __name__ == "__main__":
    unittest2.main()



# Test the symmetric code

import unittest2
from cStringIO import StringIO
import array

import chemfp
from chemfp import search, bitops

from support import fullpath, PUBCHEM_SDF, PUBCHEM_SDF_GZ

fps = chemfp.load_fingerprints(fullpath("queries.fps"))

zeros = chemfp.load_fingerprints(StringIO("""\
0000\tA
0000\tB
0001\tC
0002\tD
FFFE\tE
FFFF\tF
"""))

def slow_counts(counts, fps, threshold,
                query_start, query_end,
                target_start, target_end):
    N = len(fps)
    query_end = min(N, query_end)
    target_end = min(N, target_end)

    for row in range(query_start, query_end):
        row_fp = fps[row][1]
        for col in range(max(row+1, target_start), target_end):
            col_fp = fps[col][1]
            if bitops.byte_tanimoto(row_fp, col_fp) >= threshold:
                if row != col:
                    #print row, col, "X"
                    counts[row] += 1
                    counts[col] += 1

def slow_threshold_search(results, fps, threshold,
                          query_start, query_end,
                          target_start, target_end):
    N = len(fps)
    query_end = min(N, query_end)
    target_end = min(N, target_end)
    
    for row in range(query_start, query_end):
        row_fp = fps[row][1]
        for col in range(max(row+1, target_start), target_end):
            col_fp = fps[col][1]
            score = bitops.byte_tanimoto(row_fp, col_fp)
            if score >= threshold:
                if row != col:
                    #print row, col, "X"
                    results[row].append((col, score))
                    #results[col].append((row, score)) # only do the upper triangle


class TestCounts(unittest2.TestCase):
    #maxDiff = 0
    def test_symmetric(self):
        # query[i] always matches target[i] so x[i] will be at least one
        x = search.count_tanimoto_hits_arena(fps, fps, 0.6)

        # This only processes the upper-triangle, and not the diagonal
        y = search.count_tanimoto_hits_symmetric(fps, 0.6)

        self.assertEquals(len(x), len(y))
        for i in range(len(x)):
            self.assertEquals(x[i]-1, y[i])


    def test_zeros(self):
        y = search.count_tanimoto_hits_symmetric(zeros, 0.9)
        self.assertSequenceEqual(y, [0, 0, 0, 0, 1, 1])
        y = search.count_tanimoto_hits_symmetric(zeros, 0.001)
        self.assertSequenceEqual(y, [0, 0, 1, 2, 2, 3])


    def test_partial_counts_with_zero_threshold(self):
        threshold = 0.0
        N = len(fps)
        counts = array.array("i", [0]*N)
        search.partial_count_tanimoto_hits_symmetric(counts, fps, threshold, 1, 5, 0, N)

        expected = [0] * N
        slow_counts(expected, fps, threshold, 1, 5, 0, N)
        self.assertSequenceEqual(counts, expected)

        search.partial_count_tanimoto_hits_symmetric(counts, fps, threshold,
                                                     3, 33, 0, N)
        slow_counts(expected, fps, threshold, 3, 33, 0, N)
        self.assertSequenceEqual(counts, expected)

    def test_partial_counts_in_columns(self):
        threshold = 0.6
        N = len(fps)
        counts = array.array("i", [0]*N)
        search.partial_count_tanimoto_hits_symmetric(counts, fps, threshold, 1, 5, 0, N)

        expected = [0] * N
        slow_counts(expected, fps, threshold, 1, 5, 0, N)
        self.assertSequenceEqual(counts, expected)

        search.partial_count_tanimoto_hits_symmetric(counts, fps, threshold,
                                                     3, 33, 0, N)
        slow_counts(expected, fps, threshold, 3, 33, 0, N)
        self.assertSequenceEqual(counts, expected)
        

    def test_partial_counts_in_rows(self):
        threshold = 0.2
        N = len(fps)
        counts = array.array("i", [0]*N)
        search.partial_count_tanimoto_hits_symmetric(counts, fps, threshold, 1, 5, 9, 30)

        expected = [0] * N
        slow_counts(expected, fps, threshold, 1, 5, 9, 30)
        self.assertSequenceEqual(counts, expected)

        search.partial_count_tanimoto_hits_symmetric(counts, fps, threshold, 20, 26, 30, 50)
        slow_counts(expected, fps, threshold, 20, 26, 30, 50)
        self.assertSequenceEqual(counts, expected)

    def test_partial_counts(self):
        threshold = 0.2
        N = len(fps)
        counts = array.array("i", [0]*N)
        expected = [0] * N

        for i in range(0, N, 10):
            for j in range(i, N, 8):
                slow_counts(expected, fps, threshold, i, i+10, j, j+8)
                search.partial_count_tanimoto_hits_symmetric(counts, fps, threshold,
                                                             i, i+10, j, j+8)
                self.assertSequenceEqual(counts, expected)

        normal = search.count_tanimoto_hits_symmetric(fps, threshold)
        self.assertSequenceEqual(normal, expected)

def _compare_search_results(self, result, expected):
    q=list(result.iter_indices_and_scores())
    for i in range(len(result)):
        self.assertEquals(len(q[i]), len(expected[i]), "length of %d" % i)
        self.assertEquals(sorted(q[i]), sorted(expected[i]), "contents of %d" % i)


class TestThreshold(unittest2.TestCase):
    def test_upper_only(self):
        # query[i] always matches target[i] so x[i] will always contain i
        x = search.threshold_tanimoto_search_arena(fps, fps, 0.9)
        x = list(x)

        # This only processes the upper-triangle, and not the diagonal
        y = search.threshold_tanimoto_search_symmetric(fps, 0.9, include_lower_triangle=False)


        rows = list(row.get_indices_and_scores() for row in y)
        row_sizes = map(len, rows)
        # Move elements to the lower triangle
        for rowno, (row, row_size) in enumerate(zip(rows, row_sizes)):
            for (colno, score) in row[:row_size]:
                assert colno > rowno, (rowno, colno)
                rows[colno].append( (rowno, score) )

            # Fill in the diagonal
            row.append((rowno, 1.0))

            # Put into a consistent order
            row.sort()
            
            # Match with the NxM algorithm
            expected_row = x[rowno]
            expected_row.reorder("increasing-index")

            self.assertEquals(row, list(expected_row), rowno)

    def test_upper_and_lower(self):
        # query[i] always matches target[i] so x[i] will always contain i
        x = search.threshold_tanimoto_search_arena(fps, fps, 0.9)

        # This only processes the upper-triangle, and not the diagonal
        y = search.threshold_tanimoto_search_symmetric(fps, 0.9)

        for i, (x_row, y_row) in enumerate(zip(x, y)):
            x_row = x_row.get_indices_and_scores()
            y_row = y_row.get_indices_and_scores()
            y_row.append((i, 1.0))
            x_row.sort()
            y_row.sort()

            self.assertEquals(x_row, y_row)

    def test_partial_search_with_zero_threshold_and_no_results_offset(self):
        threshold = 0.0
        N = len(fps)
        result = search.SearchResults(4, N, fps.arena_ids)
        search.partial_threshold_tanimoto_search_symmetric(result, fps, threshold, 1, 5, 0, N)

        expected = [[] for i in range(N)]
        slow_threshold_search(expected, fps, threshold, 1, 5, 0, N)
        expected = expected[1:5]
        _compare_search_results(self, result, expected)

    def test_partial_search_with_zero_threshold_and_results_offset(self):
        threshold = 0.0
        N = len(fps)
        result = search.SearchResults(N, N, fps.arena_ids)
        search.partial_threshold_tanimoto_search_symmetric(result, fps, threshold,
                                                           1, 5, 0, N, results_offset=1)

        expected = [[] for i in range(N)]
        slow_threshold_search(expected, fps, threshold, 1, 5, 0, N)
        _compare_search_results(self, result, expected)

    def test_partial_search_in_rows(self):
        threshold = 0.2
        N = len(fps)
        result = search.SearchResults(N, N, fps.arena_ids)
        search.partial_threshold_tanimoto_search_symmetric(result, fps, threshold,
                                                           1, 9, 0, N, results_offset=1)

        expected = [[] for i in range(N)]
        slow_threshold_search(expected, fps, threshold, 1, 9, 0, N)
        _compare_search_results(self, result, expected)


    def test_partial_search_in_cols(self):
        threshold = 0.1
        N = len(fps)
        result = search.SearchResults(N, N, fps.arena_ids)
        search.partial_threshold_tanimoto_search_symmetric(result, fps, threshold,
                                                           0, N, 7, 69)

        expected = [[] for i in range(N)]
        slow_threshold_search(expected, fps, threshold, 0, N, 7, 69)
        _compare_search_results(self, result, expected)
        _compare_search_results(self, result, expected)

    
    def test_partial_threshold_search(self):
        threshold = 0.1
        N = len(fps)
        result = search.SearchResults(N, N, fps.arena_ids)
        expected = [[] for i in range(N)]

        for i in range(0, N, 13):
            for j in range(i, N, 8):
                search.partial_threshold_tanimoto_search_symmetric(result, fps, threshold,
                                                                   i, i+13, j, j+8, results_offset=i)
                slow_threshold_search(expected, fps, threshold, i, i+13, j, j+8)

        _compare_search_results(self, result, expected)

        counts_before = map(len, result)
        search.fill_lower_triangle(result)
        counts_after = map(len, result)
        self.assertNotEqual(counts_before, counts_after)
        self.assertSequenceEqual(counts_after,
                                 search.count_tanimoto_hits_symmetric(fps, threshold))

        normal = search.threshold_tanimoto_search_symmetric(fps, threshold)
        _compare_search_results(self, result, list(normal.iter_indices_and_scores()))
        
        

class TestKNearest(unittest2.TestCase):
    def test_symmetric(self):
        # query[i] always matches target[i] so x[i] will always contain element[i]
        x = search.knearest_tanimoto_search_arena(fps, fps, 31, 0.9)

        # This only processes the upper-triangle, and not the diagonal
        y = search.knearest_tanimoto_search_symmetric(fps, 30, 0.9)
        
        for i, (x_row, y_row) in enumerate(zip(x, y)):
            x_row = x_row.get_indices_and_scores()
            y_row = y_row.get_indices_and_scores()
            y_row.append((i, 1.0))
            x_row.sort()
            y_row.sort()
            self.assertEquals(x_row, y_row, "Problem in %d" % i)

    def test_symmetric2(self):
        # query[i] always matches target[i] so x[i] will always contain element[i]
        x = fps.knearest_tanimoto_search_arena(fps, 81, 0.9)

        # This only processes the upper-triangle, and not the diagonal
        y = search.knearest_tanimoto_search_symmetric(fps, 80, 0.9)

        for i, (x_row, y_row) in enumerate(zip(x, y)):
            x_row = x_row.get_indices_and_scores()
            y_row = y_row.get_indices_and_scores()
            y_row.append((i, 1.0))
            x_row.sort()
            y_row.sort()
            self.assertEquals(x_row, y_row, "Problem in %d" % i)

if __name__ == "__main__":
    unittest2.main()



# This tests that the tox environment is set up the way it should be
# set up, and that the available toolkits matches the environment.

import sys
import os
import unittest2

import support
import chemfp

envstr = os.environ.get("TOX_CHEMFP_TEST", None)
if not envstr:
    versions = []
else:
    versions = envstr.split(",")

def check_py27():
    version = sys.version_info[:2]
    assert version == (2,7), version
    assert not support.can_skip("py27")
    
def check_py35():
    version = sys.version_info[:2]
    assert version == (3,5), version
    assert not support.can_skip("py35")

### I no longer have a 32-bit test environment.
## def check_x32():
##     assert sys.maxint == 2147483647
##     assert not support.can_skip("x32")

def check_x64():
    # sys.maxint doesn't exist under Python 3.
    if sys.version_info[0] == 2:
        assert sys.maxint == 9223372036854775807
    assert not support.can_skip("x64")

## def check_oe174():
##     from openeye.oechem import OEChemGetRelease
##     version = OEChemGetRelease()
##     assert version == "1.7.4", version
##     assert not support.can_skip("oe")
##     assert not support.can_skip("oe174")
    

## def check_oe2011Oct():
##     from openeye.oechem import OEChemGetRelease
##     version = OEChemGetRelease()
##     assert version == "1.7.6", version
##     assert not support.can_skip("oe")
##     assert not support.can_skip("oe2011Oct")

## def check_oe2013Jun4():
##     from openeye.oechem import OEChemGetRelease
##     version = OEChemGetRelease()
##     assert version == "1.9.2", version
##     assert not support.can_skip("oe")
##     assert not support.can_skip("oe2013Jun4")

def check_oe20140707():
    from openeye.oechem import OEChemGetRelease
    version = OEChemGetRelease()
    assert version == "2.0.1", version
    assert not support.can_skip("oe")
    assert not support.can_skip("oe20140707")

def check_oe20161001():
    from openeye.oechem import OEChemGetRelease
    version = OEChemGetRelease()
    assert version == "2.1.0", version
    assert not support.can_skip("oe")
    assert not support.can_skip("oe20161001")

def check_oe201710b2():
    from openeye.oechem import OEChemGetRelease
    version = OEChemGetRelease()
    assert version == "2.1.3.b.2 debug", version
    assert not support.can_skip("oe")
    assert not support.can_skip("oe201710b2")


def check_ob230():
    import openbabel
    from chemfp import openbabel as _openbabel_toolkit
    assert _openbabel_toolkit._ob_version == "2.3.0", openbabel._ob_version
    assert not support.can_skip("ob")
    assert not support.can_skip("ob230")

## def check_ob23svn1():
##     import openbabel
##     from chemfp import _openbabel_toolkit
##     assert _openbabel_toolkit._ob_version == "2.3.90", openbabel._ob_version
##     assert not support.can_skip("ob")
##     assert not support.can_skip("ob23svn1")

def check_ob241():
    import openbabel
    from chemfp import openbabel as _openbabel_toolkit
    assert _openbabel_toolkit._ob_version == "2.4.1", openbabel._ob_version
    assert not support.can_skip("ob")
    assert not support.can_skip("ob241")

# It was hard ensure the pattern version were correct.
def check_rd201303():
    import rdkit.rdBase
    from rdkit.rdBase import rdkitVersion
    assert rdkitVersion[:7] == "2013.03", rdkitVersion

    from chemfp.rdkit import PATTERN_VERSION
    assert PATTERN_VERSION == "1", PATTERN_VERSION
    
def check_rd201609():
    import rdkit.rdBase
    from rdkit.rdBase import rdkitVersion
    assert rdkitVersion[:7] == "2016.09", rdkitVersion

    from chemfp.rdkit import PATTERN_VERSION
    assert PATTERN_VERSION == "2", PATTERN_VERSION
    
def check_rd201703():
    import rdkit.rdBase
    from rdkit.rdBase import rdkitVersion
    assert rdkitVersion == "2017.03.3", rdkitVersion

    from chemfp.rdkit import PATTERN_VERSION
    assert PATTERN_VERSION == "3", PATTERN_VERSION
    
def check_rd201709_dev():
    import rdkit.rdBase
    from rdkit.rdBase import rdkitVersion
    assert rdkitVersion == "2017.09.1.dev1", rdkitVersion

    from chemfp.rdkit import PATTERN_VERSION
    assert PATTERN_VERSION == "4", PATTERN_VERSION
    

def _check(required):
    req = required.split()
    for name in versions:
        if name in req:
            return
    raise AssertionError("Missing one of %r: %r" % (required, versions))

class TestToxVersion(unittest2.TestCase):
    def test_enough_specifications(self):
        _check("x64")
        _check("py27 py35")
        _check("ob241 rd201303 rd201609 oe20140707 oe20161001 oe201710b2 rd201703 rd201709_dev")
        
    def test_version(self):
        for name in versions:
            func = globals()["check_" + name]
            func()

TestToxVersion = unittest2.skipUnless(envstr, "Not building under the tox environment")(
    TestToxVersion)

if __name__ == "__main__":
    unittest2.main()



# Copyright (c) 2011-2012 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "__init__.py" for full license details.

import heapq
import operator

from . import bitops
from . import Fingerprints
from . import fps_search

def count_tanimoto_hits_fp(query_fp, targets, threshold):
    return sum(1 for target in targets
                   if bitops.byte_tanimoto(query_fp, target[1]) >= threshold)

## def iter_count_tanimoto_hits(queries, targets, threshold):
##     for query_id, query_fp in queries:
##         yield query_id, sum(1 for target in targets
##                                if bitops.byte_tanimoto(query_fp, target[1]) >= threshold)

def count_tanimoto_hits_arena(queries, targets, threshold):
    counts = []
    for query_id, query_fp in queries:
        counts.append(sum(1 for target in targets
                          if bitops.byte_tanimoto(query_fp, target[1]) >= threshold))
    return counts

def tanimoto_count_once(queries, targets, threshold):
    # Only go through the list of queries once
    results = [0] * len(queries)
    query_ids = []
    query_fps = []
    for (id, fp) in queries:
        query_ids.append(id)
        query_fps.append(fp)
    for target_id, target_fp in targets:
        for i, query_fp in enumerate(query_fps):
            if bitops.byte_tanimoto(query_fp, target_fp) >= threshold:
                results[i] += 1
    return zip(query_ids, results)

##########

def threshold_tanimoto_search_fp(query_fp, targets, threshold):
    ids = []
    scores = []
    for target_id, target_fp in targets:
        score = bitops.byte_tanimoto(query_fp, target_fp)
        if score >= threshold:
            ids.append(target_id)
            scores.append(score)
    return fps_search.FPSSearchResult(ids, scores)

def _iter_threshold_tanimoto_search(queries, targets, threshold):
    for query_id, query_fp in queries:
        yield threshold_tanimoto_search_fp(query_fp, targets, threshold)

def threshold_tanimoto_search_arena(queries, targets, threshold):
    results = []
    for query_id, query_fp in queries:
        results.append(threshold_tanimoto_search_fp(query_fp, targets, threshold))
    return fps_search.FPSSearchResults(results)

#def tanimoto_search_all(queries, targets, threshold):
#    results = [[] for i in xrange(len(queries))]
#    query_fps = [query[1] for query in queries]
#    for target_id, target_fp in targets:
#        for i, query_fp in enumerate(query_fps):
#            score = bitops.byte_tanimoto(query_fp, target_fp)
#            if score >= threshold:
#                results[i].append( (target_id, score) )
#    return results

##########


def knearest_tanimoto_search_fp(query_fp, targets, k, threshold):
    hits = heapq.nlargest(k, threshold_tanimoto_search_fp(query_fp, targets, threshold),
                          key = operator.itemgetter(1))
    if hits:
        ids, scores = zip(*hits)
    else:
        ids, scores = [], []
    return fps_search.FPSSearchResult(ids, scores)

def _iter_knearest_tanimoto_search(queries, targets, k, threshold):
    for hits in _iter_threshold_tanimoto_search(queries, targets, threshold):
        hits = heapq.nlargest(k, hits, key = operator.itemgetter(1))
        ids, scores = zip(*hits)
        yield fps_search.FPSSearchResult(ids, scores)

def knearest_tanimoto_search_arena(queries, targets, k, threshold):
    results = []
    for row in _iter_knearest_tanimoto_search(queries, targets, k, threshold):
        results.append(row)
    return fps_search.FPSSearchResults(results)

# I am not going to optimize this.
#def knearest_tanimoto_search_all(queries, targets, k, threshold):
#    results = tanimoto_search(queries, targets, threshold)
#    for i, (id, hits) in enumerate(results):
#        results[i] = (id, heapq.nlargest(k, hits, key = operator.itemgetter(1)))
#    return results

def _check_threshold(threshold):
    if not (0.0 <= threshold <= 1.0):
        raise ValueError("threshold must between 0.0 and 1.0, inclusive")

class SlowFingerprints(Fingerprints):
    def count_tanimoto_hits_fp(self, query_fp, threshold=0.7):
        if not (0.0 <= threshold <= 1.0):
            raise ValueError("threshold must between 0.0 and 1.0, inclusive")
        return count_tanimoto_hits_fp(query_fp, self._id_fp_pairs, threshold)

    def count_tanimoto_hits_arena(self, queries, threshold=0.7):
        _check_threshold(threshold)
        return count_tanimoto_hits_arena(queries, self._id_fp_pairs, threshold)

    def threshold_tanimoto_search_fp(self, fp, threshold=0.7):
        _check_threshold(threshold)
        return threshold_tanimoto_search_fp(fp, self._id_fp_pairs, threshold)

    def threshold_tanimoto_search_arena(self, queries, threshold=0.7):
        _check_threshold(threshold)
        return threshold_tanimoto_search_arena(queries, self._id_fp_pairs, threshold)

    def knearest_tanimoto_search_fp(self, fp, k=3, threshold=0.7):
        if k < 0:
            raise ValueError("k must be non-negative")
        _check_threshold(threshold)
        return knearest_tanimoto_search_fp(fp, self._id_fp_pairs, k, threshold)

    def knearest_tanimoto_search_arena(self, queries, k=3, threshold=0.7, batch_size=100):
        if k < 0:
            raise ValueError("k must be non-negative")
        _check_threshold(threshold)
        return knearest_tanimoto_search_arena(queries, self._id_fp_pairs, k, threshold)



from __future__ import absolute_import

# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

import os
import sys
import _chemfp
from _chemfp import (hex_isvalid, hex_popcount, hex_intersect_popcount,
                     hex_tanimoto, hex_contains, hex_contains_bit,
                     hex_encode, hex_encode_as_bytes, hex_decode,
                     hex_from_bitlist, hex_to_bitlist,
                     hex_intersect, hex_union, hex_difference)
                    

from _chemfp import (byte_popcount, byte_intersect_popcount,
                     byte_tanimoto, byte_contains, byte_contains_bit,
                     byte_hex_tanimoto,
                     byte_from_bitlist, byte_to_bitlist,
                     byte_intersect, byte_union, byte_difference)

__all__ = ["byte_popcount", "byte_intersect_popcount",
           "byte_tanimoto", "byte_contains", "byte_contains_bit",
           "byte_to_bitlist", "byte_from_bitlist",
           "hex_isvalid", "hex_popcount", "hex_intersect_popcount", 
           "hex_tanimoto", "hex_contains",
           "hex_to_bitlist", "hex_from_bitlist",
           "hex_encode", "hex_encode_as_bytes", "hex_decode"
           # Don't export the method and alignment-related functions
               ]


def get_methods():
    return [_chemfp.get_method_name(i) for i in range(_chemfp.get_num_methods())]

def get_alignments():
    return [_chemfp.get_alignment_name(i) for i in range(_chemfp.get_num_alignments())]

def get_alignment_methods():
    settings = {}
    for alignment in range(_chemfp.get_num_alignments()):
        method = _chemfp.get_alignment_method(alignment)
        settings[_chemfp.get_alignment_name(alignment)] = _chemfp.get_method_name(method)
    return settings

def get_alignment_method(alignment):
    try:
        alignment_i = get_alignments().index(alignment)
    except ValueError:
        raise ValueError("Unknown alignment %r" % (alignment,))
    return _chemfp.get_method_name(_chemfp.get_alignment_method(alignment_i))

def set_alignment_method(alignment, method):
    try:
        alignment_i = get_alignments().index(alignment)
    except ValueError:
        raise ValueError("Unknown alignment %r" % (alignment,))

    try:
        method_i = get_methods().index(method)
    except ValueError:
        raise ValueError("Unknown method %r" % (method,))

    _chemfp.set_alignment_method(alignment_i, method_i)

def select_fastest_method(repeat=10000):
    if repeat > 100000:
        raise ValueError("repeat size is meaninglessly large")
    if repeat < 1:
        raise ValueError("repeat size must be 1 or larger (values under 1000 are likely useless)")
    
    for alignment_i, name in enumerate(get_alignments()):
        _chemfp.select_fastest_method(alignment_i, repeat)


def get_options():
    return [_chemfp.get_option_name(i) for i in range(_chemfp.get_num_options())]

def get_option(option):
    _chemfp.get_option(option)

def set_option(option, value):
    _chemfp.set_option(option, value)

def print_report(out=sys.stdout):
    from . import SOFTWARE
    print >>out, "== Configuration report for", SOFTWARE, "=="
    print >>out, "Number of available threads:", _chemfp.get_num_threads()
    print >>out, "Available methods:", " ".join(get_methods())
    print >>out, "Alignment methods:"
    for alignment in get_alignments():
        method = get_alignment_method(alignment)
        print >>out, "  %s: %s" % (alignment, method)
    print >>out, "Option settings:"
    for option in get_options():
        print >>out, "  %s: %s" % (option, _chemfp.get_option(option))

def use_environment_variables(environ=None):
    if environ is None:
        environ = os.environ

    known = set()
    for alignment in get_alignments():
        name = "CHEMFP_" + alignment.upper().replace("-", "_")
        known.add(name)
        try:
            value = environ[name]
        except KeyError:
            pass
        else:
            try:
                set_alignment_method(alignment, value)
            except ValueError, err:
                print >>sys.stderr, "WARNING: Unable to use $%s = %r: %s" % (
                    (name, value, err))

    for option in get_options():
        name = "CHEMFP_" + option.upper().replace("-", "_")
        known.add(name)
        try:
            value = environ[name]
        except KeyError:
            pass
        else:
            try:
                set_option(option, int(value))
            except ValueError, err:
                print >>sys.stderr, "WARNING: Unable to use $%s = %r: %s" % (
                    (name, value, err))

    known.add("CHEMFP_REPORT")
    report = environ.get("CHEMFP_REPORT", "0") == "1"

    if report:
        set_option("report-popcount", 1)
        set_option("report-intersect", 1)

    
    known.add("CHEMFP_PRINT_CONFIG")
    if (environ.get("CHEMFP_PRINT_CONFIG", "0") == "1" or report):
        print_report(sys.stderr)


    for k in environ:
        if not k.startswith("CHEMFP_"):
            continue
        if k not in known:
            print >>sys.stderr, "WARNING: Unknown chemfp environment variable %r" % (k,)



from __future__ import with_statement
import sys
import unittest2
import support

try:
    from openeye.oechem import OEGraphMol, OEParseSmiles, OEChemIsLicensed
    if not OEChemIsLicensed():
        raise ImportError
    skip_oechem = False
except ImportError:
    skip_oechem = support.can_skip("oe")

if not skip_oechem:
    from chemfp import openeye_patterns

import test_patterns
from chemfp import types

MACCS_SMI = support.fullpath("maccs.smi")

def parse_smiles(smiles):
    mol = OEGraphMol()
    OEParseSmiles(mol, smiles)
    return mol

def _count(it):
    return sum(1 for item in it)

class ReferenceMixin(object):
    def test_reference_data_set(self):
        largest = min(self.reference_limit, max(v for (k,v) in self.reference_cases))

        matcher = self.reference_class(largest)
            
        for (smiles, expected) in self.reference_cases:
            mol = parse_smiles(smiles)

            self.assertEquals(matcher.SingleMatch(mol), bool(expected), smiles)

            expected = min(expected, largest)
            self.assertGreaterEqual(_count(matcher.Match(mol)), expected, smiles)

    def test_match_limit(self):
        largest = min(5, self.reference_limit)
        for max_count in range(1, largest+1):
            matcher = self.reference_class(max_count)
            for (smiles, expected) in self.reference_cases:
                mol = parse_smiles(smiles)
                expected = min(expected, max_count)
                self.assertGreaterEqual(_count(matcher.Match(mol)), expected, smiles)

class TestHydrogenMatcher(ReferenceMixin, unittest2.TestCase):
    if not skip_oechem:
        reference_class = openeye_patterns.HydrogenMatcher
        reference_cases = test_patterns.hydrogen_test_cases
        reference_limit = 100

TestHydrogenMatcher = unittest2.skipIf(skip_oechem, "OEChem not installed")(
    TestHydrogenMatcher)

class TestAromaticRingMatcher(ReferenceMixin, unittest2.TestCase):
    if not skip_oechem:
        reference_class = openeye_patterns.AromaticRings
        reference_cases = test_patterns.aromatic_ring_cases
        reference_limit = 2

TestAromaticRingMatcher = unittest2.skipIf(skip_oechem, "OEChem not installed")(
    TestAromaticRingMatcher)

# XXX These are too low-level. The tests should really be done through the
# file interface and shared with the other implementations.
 
class TestAromaticRings(unittest2.TestCase):
    def _count(self, matcher, smiles):
        return sum(1 for x in matcher.Match(parse_smiles(smiles)))
    def test_at_least_one_aromatic_ring(self):
        matcher = openeye_patterns.AromaticRings(1)
        self.assertFalse(matcher.SingleMatch(parse_smiles("C")))
        self.assertTrue(matcher.SingleMatch(parse_smiles("c1ccccc1")))
        self.assertTrue(matcher.SingleMatch(parse_smiles("c1ccccc1c1ccccc1")))

        self.assertEqual(self._count(matcher, "C"), 0)
        self.assertEqual(self._count(matcher, "c1ccccc1"), 1)
        self.assertIn(self._count(matcher, "c12cccccc1ccccc2"), (1, 2)) # Matcher can return >count; XXX Why?
        self.assertEqual(self._count(matcher, "c1ccccc1c1ccccc1"), 1)
        self.assertEqual(self._count(matcher, "c1ccccc1C1CCCCC1"), 1)

    def test_at_least_two_aromatic_rings(self):
        matcher = openeye_patterns.AromaticRings(2)

        self.assertEqual(self._count(matcher, "C"), 0)
        self.assertEqual(self._count(matcher, "c1ccccc1"), 1)
        self.assertEqual(self._count(matcher, "c12cccccc1ccccc2"), 2)
        self.assertEqual(self._count(matcher, "c1ccccc1c1ccccc1"), 2)
        self.assertEqual(self._count(matcher, "c1ccccc1C1CCCCC1"), 1)
        self.assertGreaterEqual(self._count(matcher, "c1ccc-2cccc-2cc1"), 1)
        self.assertGreaterEqual(self._count(matcher, "c1ccc-2cccc-2c3c1cccn3"), 2)
        self.assertGreaterEqual(self._count(matcher, "c1ccccc1.c1ccccc1"), 2)

    def test_aromatic_rings_failure(self):
        with self.assertRaisesRegexp(NotImplementedError, "No support for >=3 aromatic rings"):
            openeye_patterns.AromaticRings(3)

class TestHeteroAromaticRings(unittest2.TestCase):
    def _count(self, matcher, smiles):
        return sum(1 for x in matcher.Match(parse_smiles(smiles)))
    def test_at_least_one_heteroaromatic_ring(self):
        matcher = openeye_patterns.HeteroAromaticRings(1)
        self.assertFalse(matcher.SingleMatch(parse_smiles("C")))
        self.assertTrue(matcher.SingleMatch(parse_smiles("c1cccnc1")))
        self.assertTrue(matcher.SingleMatch(parse_smiles("c1cccnc1c1ccccc1")))
        self.assertTrue(matcher.SingleMatch(parse_smiles("c1ccccc1c1ncccc1")))

        self.assertEqual(self._count(matcher, "C"), 0)
        self.assertEqual(self._count(matcher, "c1ccccc1"), 0)
        self.assertEqual(self._count(matcher, "c1ccncc1"), 1)
        self.assertEqual(self._count(matcher, "c12cccccc1ccccc2"), 0)
        self.assertEqual(self._count(matcher, "c12ccnc1ccccc2"), 1)
        self.assertEqual(self._count(matcher, "c12cncc1ccccc2"), 1)
        self.assertEqual(self._count(matcher, "c12ccccc1cncc2"), 1)
        self.assertIn(self._count(matcher, "c12cccnc1cncc2"), (1, 2))  # a matcher can return >count. Why? XXX
        self.assertEqual(self._count(matcher, "c1ccccc1c1ccccc1"), 0)
        self.assertEqual(self._count(matcher, "c1ccccc1C1CCCNC1"), 0)
        self.assertEqual(self._count(matcher, "c1cnccc1C1CCCNC1"), 1)
        self.assertEqual(self._count(matcher, "c1cnccc1C1CCCNC1"), 1)

    def test_at_least_two_heteroaromatic_rings(self):
        matcher = openeye_patterns.HeteroAromaticRings(2)
        
        self.assertEqual(self._count(matcher, "C"), 0)
        self.assertEqual(self._count(matcher, "c1ccccc1"), 0)
        self.assertEqual(self._count(matcher, "c1ccncc1"), 1)
        self.assertEqual(self._count(matcher, "c12cccccc1ccccc2"), 0)
        self.assertEqual(self._count(matcher, "c12ccnc1ccccc2"), 1)
        self.assertEqual(self._count(matcher, "c12cncc1ccccc2"), 1)
        self.assertEqual(self._count(matcher, "c12ccccc1cncc2"), 1)
        self.assertEqual(self._count(matcher, "c12cccnc1cncc2"), 2)
        self.assertEqual(self._count(matcher, "c1ccccc1c1ccccc1"), 0)
        self.assertEqual(self._count(matcher, "c1ccccc1C1CCCNC1"), 0)
        self.assertEqual(self._count(matcher, "c1cnccc1C1CCCNC1"), 1)
        self.assertEqual(self._count(matcher, "c1cnccc1C1CCCNC1"), 1)

        self.assertEqual(self._count(matcher, "c1ccccn1.c1ccoc1"), 2)
        self.assertEqual(self._count(matcher, "c1ccc-2cccc-2cc1"), 0)
        self.assertEqual(self._count(matcher, "c1ccn-2ccc-2cc1"), 1)
        self.assertEqual(self._count(matcher, "c1ccccc2c1cncnc2"), 1)

    def test_heteroaromatic_rings_failure(self):
        with self.assertRaisesRegexp(NotImplementedError, "No support for >=3 hetero-aromatic rings"):
            openeye_patterns.HeteroAromaticRings(3)


class TestNumFragments(unittest2.TestCase):
    def _count(self, matcher, smiles):
        return sum(1 for x in matcher.Match(parse_smiles(smiles)))
    def test_single_match(self):
        matcher = openeye_patterns.NumFragments(10)
        self.assertEqual(matcher.SingleMatch(OEGraphMol()), 0)
        self.assertEqual(matcher.SingleMatch(parse_smiles("C")), 1)
        self.assertEqual(matcher.SingleMatch(parse_smiles("C.C")), 1)

    def test_multiple_matches(self):
        matcher = openeye_patterns.NumFragments(10)
        self.assertEqual(self._count(matcher, ""), 0)
        self.assertEqual(self._count(matcher, "C"), 1)
        self.assertEqual(self._count(matcher, "CC"), 1)
        self.assertEqual(self._count(matcher, "C.C"), 2)
        self.assertEqual(self._count(matcher, "C.C.O.[U]"), 4)

TestAromaticRings = unittest2.skipIf(skip_oechem, "OEChem not installed")(TestAromaticRings)
TestHeteroAromaticRings = unittest2.skipIf(skip_oechem, "OEChem not installed")(TestHeteroAromaticRings)
TestNumFragments = unittest2.skipIf(skip_oechem, "OEChem not installed")(TestNumFragments)

if __name__ == "__main__":
    unittest2.main()



from __future__ import print_function

# Copyright (c) 2010-2017 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

import os

class UnsupportedPatternError(KeyError):
    def __init__(self, pattern, reason=None):
        KeyError.__init__(self, pattern)
        self.pattern = pattern
        if reason is None:
            reason = "Cannot interpret pattern definition"
        self.reason = reason
        self.filename = None
        self.lineno = None
    def __str__(self):
        msg = self.reason  + " " + repr(self.pattern)
        if self.lineno is not None:
            msg += " at line %d" % (self.lineno,)
        if self.filename is not None:
            msg += " in file %r" % (self.filename,)
        return msg

class PatternFile(object):
    def __init__(self, filename, max_bit, bit_definitions):
        assert max_bit >= 0, max_bit
        self.filename = filename
        self.max_bit = max_bit
        self.bit_definitions = bit_definitions
        self._bit_to_bit_definition = dict((bitdef.bit, bitdef) for bitdef in bit_definitions)

    def __getitem__(self, bit):
        return self._bit_to_bit_definition[bit]

    def __iter__(self):
        return iter(self._bit_to_bit_definition)

    
class BitDefinition(object):
    __slots__ = ("bit", "count", "pattern", "description", "lineno")
    def __init__(self, bit, count, pattern, description, lineno):
        self.bit = bit
        self.count = count
        self.pattern = pattern
        self.description = description
        self.lineno = lineno
    

def load_patterns(infile):
    if isinstance(infile, basestring):
        infile = open(infile, "rU")
    filename = getattr(infile, "name", "<unknown>")
    bit_definitions = list(read_patterns(infile))
    max_bit = max(bitdef.bit for bitdef in bit_definitions)
    
    return PatternFile(filename, max_bit, bit_definitions)

def read_patterns(infile):
    seen_bits = {}

    for lineno, line in enumerate(infile):
        lineno += 1

        # Leading and trailing whitespace is ignored
        line = line.strip()
        
        # Ignore blank lines or those with a leading "#"
        if not line or line.startswith("#"):
            continue

        # The first three columns, plus everything else for the description
        fields = line.split(None, 3)
        if len(fields) != 4:
            raise TypeError("Not enough fields on line %d: %r" % (lineno, line))

        # Normalize whitespace for the description
        fields[3] = " ".join(fields[3].split())

        # Do some type checking and error reporting
        bit, count, pattern, description = fields
        if not bit.isdigit():
            raise TypeError(
                "First field of line %d must be a non-negative bit position, not %r" %
                            (lineno, bit))
        bit = int(bit)

        if not count.isdigit() or int(count) == 0:
            raise TypeError(
                "Second field of line %d must be a positive minimum match count, not %r" %
                            (lineno, bit))
        count = int(count)
        
        if bit in seen_bits:
            raise TypeError("Line %d redefines bit %d, already set by line %d" %
                            (lineno, bit, seen_bits[bit]))
        seen_bits[bit] = lineno

        yield BitDefinition(bit, count, pattern, description, lineno)

class CountInfo(object):
    __slots__ = ("count", "bit", "byteno", "bitmask")
    def __init__(self, count, bit):
        self.count = count  # minimum count needed to enable this bit
        self.bit = bit  # used to set not_implemented, and useful for debugging
        
        # These simplify the fingerprint generation code
        self.byteno = bit//8
        self.bitmask = 1<<(bit%8)


def _bit_definition_to_pattern_definition(bit_definitions):
    "Helper function to organize the bit defintions based on pattern instead of bit"
    
    # A pattern definition is of the form:
    #  (pattern string, count_info_list)
    #     where the count_info list elements are sorted by count

    # I want to preserve the pattern order so that patterns which
    # are defined first are evaluated first
    ordered_patterns = []
    pattern_info = {}

    # Find all of the bit definitions for a given pattern
    for bitdef in bit_definitions:
        if bitdef.pattern not in pattern_info:
            pattern_info[bitdef.pattern] = []
            ordered_patterns.append(bitdef.pattern)
        pattern_info[bitdef.pattern].append( CountInfo(bitdef.count, bitdef.bit) )

    # Put them into a slighly more useful form
    #  - sorted now makes it easier to test when done
    #  - knowing the max match count lets some matchers optmize how to match
    for pattern in ordered_patterns:
        count_info_list = pattern_info[pattern]
        count_info_list.sort(key=lambda count_info: count_info.count)
        yield (pattern,
               count_info_list[-1].count,  # the largest count
               tuple(count_info_list)
               )

class LimitedMatcher(object):
    def __init__(self, max_supported, matcher):
        self.max_supported = max_supported
        self.matcher = matcher
                 
def _build_matchers(patterns, pattern_definitions, compile_pattern):
    not_implemented = set()
    matcher_definitions = []
    for (pattern, largest_count, count_info_tuple) in pattern_definitions:
        if pattern == "<0>":
            # Special case support for setting (or rather, ignoring) the 0 bit
            continue

        matcher = compile_pattern(pattern, largest_count)
        if isinstance(matcher, LimitedMatcher):
            max_supported = matcher.max_supported
            new_count_info = []
            for count_info in count_info_tuple:
                if count_info.count <= max_supported:
                    new_count_info.append(count_info)
                else:
                    not_implemented.add(count_info.bit)
            
            matcher = matcher.matcher
            count_info_tuple = tuple(new_count_info)
            if not count_info_tuple:
                continue

        if matcher is None:
            # During development I sometimes forgot to return a matcher
            # This catches those cases
            raise UnsupportedPatternError(pattern)
        
        matcher_definitions.append( (matcher, largest_count, count_info_tuple) )

    return not_implemented, tuple(matcher_definitions)

def make_matchers(patterns, compile_pattern):
    pattern_definitions = _bit_definition_to_pattern_definition(patterns.bit_definitions)
    try:
        return _build_matchers(patterns, pattern_definitions, compile_pattern)
    except UnsupportedPatternError, err:
        err.filename = patterns.filename
        
        pattern = err.args[0]
        for bitdef in patterns.bit_definitions:
            if bitdef.pattern == pattern:
                err.lineno = bitdef.lineno
                raise
        raise
        

class PatternFingerprinter(object):
    def __init__(self, patterns, compile_pattern):
        self.patterns = patterns

        self.num_bytes = (patterns.max_bit // 8) + 1
        self.not_implemented, self.matcher_definitions = (
            make_matchers(patterns, compile_pattern)   )

    def describe(self, bit):
        description = self.patterns[bit].description
        if bit in self.not_implemented:
             description + " (NOT IMPLEMENTED)"
        return description

    def fingerprint(self, mol):
        raise NotImplemented("Must be implemented by a derived class")


def _load_named_patterns(name):
    filename = os.path.join(os.path.dirname(__file__), name + ".patterns")
    return load_patterns(filename)



from __future__ import with_statement

from cStringIO import StringIO

import unittest2
import ctypes

import chemfp
from chemfp import arena, bitops
import _chemfp

zeros = ("0000\tfirst\n"
         "0010\tsecond\n"
         "0000\tthird\n")

ordered_zeros = ("0000\tfirst\n"
                 "0000\tthird\n"
                 "0010\tsecond\n")

class TestUnsortedAlignment(unittest2.TestCase):

    # Python strings are 1-, 2-, and 4-byte aligned
    
    def test_1_alignment(self):
        a = chemfp.load_fingerprints(StringIO(zeros), reorder=False, alignment=1)
        self.assertEquals(a.start_padding, 0)
        self.assertEquals(a.end_padding, 0)
        self.assertEquals(a.arena, "\x00\x00\x00\x10\x00\x00")
        self.assertEquals(a.storage_size, 2)
    def test_2_alignment(self):
        a = chemfp.load_fingerprints(StringIO(zeros), reorder=False, alignment=2)
        self.assertEquals(a.start_padding, 0)
        self.assertEquals(a.end_padding, 0)
        self.assertEquals(a.arena, "\x00\x00\x00\x10\x00\x00")
        self.assertEquals(a.storage_size, 2)
    def test_4_alignment(self):
        a = chemfp.load_fingerprints(StringIO(zeros), reorder=False, alignment=4)
        self.assertEquals(a.start_padding, 0)
        self.assertEquals(a.end_padding, 0)
        self.assertEquals(a.arena, "\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00")
        self.assertEquals(a.storage_size, 4)

    # A Python string might be 8-aligned, but no guarantee
    def test_8_alignment(self):
        arenas = [chemfp.load_fingerprints(StringIO(zeros), reorder=False, alignment=8)
                  for i in range(10)]
        for a in arenas:
            if a.start_padding == a.end_padding == 0:
                s = a.arena
            else:
                self.assertEquals(a.arena[:a.start_padding], "\x00" * a.start_padding)
                self.assertEquals(a.arena[-a.end_padding:], "\x00" * a.end_padding)
                s = a.arena[a.start_padding:-a.end_padding]
            
            self.assertEquals(s,
                              "\x00\x00\x00\x00\x00\x00\x00\x00"
                              "\x00\x10\x00\x00\x00\x00\x00\x00"
                              "\x00\x00\x00\x00\x00\x00\x00\x00")
            self.assertEquals(a.storage_size, 8)

class TestReorderedAlignment(unittest2.TestCase):
    def test_1_alignment(self):
        a = chemfp.load_fingerprints(StringIO(zeros), reorder=True, alignment=1)
        self.assertEquals(a.start_padding, 0)
        self.assertEquals(a.end_padding, 0)
        self.assertEquals(a.arena, "\x00\x00\x00\x00\x00\x10")

    def test_2_alignment(self):
        a = chemfp.load_fingerprints(StringIO(zeros), reorder=True, alignment=2)
        self.assertEquals(a.start_padding, 0)
        self.assertEquals(a.end_padding, 1)
        # The code overallocates one byte
        self.assertEquals(a.arena, "\x00\x00\x00\x00\x00\x10\x00")

    def test_4_alignment(self):
        a = chemfp.load_fingerprints(StringIO(zeros), reorder=True, alignment=4)
        self.assertEquals(a.start_padding, 0)
        self.assertEquals(a.end_padding, 3)
        self.assertEquals(a.arena,
                          "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00")


class TestAlreadyOrderedAlignment(unittest2.TestCase):
    def test_1_alignment(self):
        a = chemfp.load_fingerprints(StringIO(ordered_zeros), reorder=True, alignment=1)
        self.assertEquals(a.start_padding, 0)
        self.assertEquals(a.end_padding, 0)
        self.assertEquals(a.arena, "\x00\x00\x00\x00\x00\x10")

    def test_2_alignment(self):
        a = chemfp.load_fingerprints(StringIO(ordered_zeros), reorder=True, alignment=2)
        self.assertEquals(a.start_padding, 0)
        self.assertEquals(a.end_padding, 0)
        self.assertEquals(a.arena, "\x00\x00\x00\x00\x00\x10")

    def test_4_alignment(self):
        a = chemfp.load_fingerprints(StringIO(ordered_zeros), reorder=True, alignment=4)
        self.assertEquals(a.start_padding, 0)
        self.assertEquals(a.end_padding, 0)
        self.assertEquals(a.arena, "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00")

    def test_16_alignment(self):
        arenas = [chemfp.load_fingerprints(StringIO(ordered_zeros), reorder=True, alignment=16)
                  for i in range(10)]
        for a in arenas:
            if a.start_padding == 0 and a.end_padding == 0:
                s = a.arena
            else:
                self.assertEquals(a.start_padding + a.end_padding + 1, 16)
                self.assertEquals(a.arena[:a.start_padding], "\x00" * a.start_padding)
                self.assertEquals(a.arena[-a.end_padding:], "\x00" * a.end_padding)
                s = a.arena[a.start_padding:-a.end_padding]

            self.assertEquals(s, ("\x00"*16 +
                                  "\x00"*16 +
                                  "\x00\x10" + "\x00"*14))

class TestOptimalAlignment(unittest2.TestCase):
    def setUp(self):
        self._has_popcnt = arena._has_popcnt
        self._has_ssse3 = arena._has_ssse3
        self.get_alignment_method = bitops.get_alignment_method
        bitops.get_alignment_method = self.hook

    def tearDown(self):
        arena._has_popcnt = self._has_popcnt
        arena._has_ssse3 = self._has_ssse3
        bitops.get_alignment_method = self.get_alignment_method

    def hook(self, name):
        return self.data[name]

    def test_tiny(self):
        for i in range(1, 9):
            self.assertEquals(arena.get_optimal_alignment(i), 1)

    def test_small(self):
        for i in range(9, 33):
            self.assertEquals(arena.get_optimal_alignment(i), 4)


    def test_medium(self):
        arena._has_popcnt = False
        arena._has_ssse3 = False
        for i in range(35, 225):
            self.assertEquals(arena.get_optimal_alignment(i), 8)

    def test_popcnt_ssse3_combinations(self):
        arena._has_popcnt = True
        arena._has_ssse3 = True
        
        self.data = {"align8-large": "POPCNT",
                     "align8-small": "POPCNT",
                     "align-ssse3": "ssse3"}
        self.assertEquals(arena.get_optimal_alignment(300), 8)
        self.assertEquals(arena.get_optimal_alignment(800), 8)
        
        self.data = {"align8-large": "POPCNT",
                     "align8-small": "LUT8-1",
                     "align-ssse3": "ssse3"}
        self.assertEquals(arena.get_optimal_alignment(300), 64)
        self.assertEquals(arena.get_optimal_alignment(800), 8)

        self.data = {"align8-large": "LUT8-1",
                     "align8-small": "POPCNT",
                     "align-ssse3": "ssse3"}
        self.assertEquals(arena.get_optimal_alignment(300), 8)
        self.assertEquals(arena.get_optimal_alignment(800), 64)

        self.data = {"align8-large": "LUT8-1",
                     "align8-small": "LUT8-1",
                     "align-ssse3": "ssse3"}
        self.assertEquals(arena.get_optimal_alignment(300), 64)
        self.assertEquals(arena.get_optimal_alignment(800), 64)

        # And now, with SSSE3 disabled

        self.data = {"align8-large": "POPCNT",
                     "align8-small": "POPCNT",
                     "align-ssse3": "POPCNT"}
        self.assertEquals(arena.get_optimal_alignment(300), 8)
        self.assertEquals(arena.get_optimal_alignment(800), 8)
        
        self.data = {"align8-large": "POPCNT",
                     "align8-small": "LUT8-1",
                     "align-ssse3": "POPCNT"}
        self.assertEquals(arena.get_optimal_alignment(300), 8)
        self.assertEquals(arena.get_optimal_alignment(800), 8)

        self.data = {"align8-large": "LUT8-1",
                     "align8-small": "POPCNT",
                     "align-ssse3": "POPCNT"}
        self.assertEquals(arena.get_optimal_alignment(300), 8)
        self.assertEquals(arena.get_optimal_alignment(800), 8)

        self.data = {"align8-large": "LUT8-1",
                     "align8-small": "LUT8-1",
                     "align-ssse3": "POPCNT"}
        self.assertEquals(arena.get_optimal_alignment(300), 8)
        self.assertEquals(arena.get_optimal_alignment(800), 8)


# I can't find a better solution than this. (!?)
def _addressof(s):
    t = str(ctypes.c_char_p(s))
    after_open_paren = t.split("(")[1]
    return int(after_open_paren.strip(")"))

class TestFingerprintAlignment(unittest2.TestCase):
    def test_different_cases(self):
        for query in (
            ("1", 4, 8),
            ("12", 8, 8),
            ("123", 16, 16),
            ("abcd", 4, 8),
            ("abcd", 8, 8),
            ("abcd", 16, 16),
            ):
            fp, alignment, storage_size = query
            
            result = _chemfp.align_fingerprint(*query)
            start_padding, end_padding, s = result
            i = _addressof(s) + start_padding
            self.assertEquals(i % alignment, 0, (query, result))

            expected = fp + "\0" * (storage_size - len(fp))
            self.assertEquals(s[start_padding:-end_padding], expected,
                              (query, expected, result))

            self.assertEquals(s[:start_padding], "\0"*start_padding)
            self.assertEquals(s[-end_padding:], "\0"*end_padding)

    def test_identical(self):
        # This fingerprint is aligned; no need to create a new one
        s = "blah"
        start_padding, end_padding, t = _chemfp.align_fingerprint(s, 4, 4)
        self.assertEquals(start_padding, 0)
        self.assertEquals(end_padding, 0)
        self.assertIs(s, t)
        
    def test_errors(self):
        with self.assertRaisesRegexp(ValueError, "must be a character buffer"):
            _chemfp.align_fingerprint(1, 4, 4)
            
        with self.assertRaisesRegexp(ValueError, "storage size is too small"):
            _chemfp.align_fingerprint("too long", 4, 4)

        with self.assertRaisesRegexp(ValueError, "storage size must be positive"):
            _chemfp.align_fingerprint("", 1, 0)

        with self.assertRaisesRegexp(ValueError, "storage size must be positive"):
            _chemfp.align_fingerprint("X", 1, -12)
            
        with self.assertRaisesRegexp(ValueError, "alignment must be a positive power of two"):
            _chemfp.align_fingerprint("1234", 3, 4)

if __name__ == "__main__":
    unittest2.main()
